;
;   logint [-b <word>] <hexno>      : installs logint for interrupt <hexno>
;   logint -u<hexno><filename>      : uninstalls logint, and saves buffer to file
;
; example :
;   logint 21
;   logint -u 21 int21.log
;
; future: configurable buffer size, and buffer located at end of resident code
;  - bug: buffer is allocated over startup code, if an int occurs while
;         still in startup process program may crash
;
farptr STRUC
  o DW ?
  s DW ?
farptr ENDS

.model tiny
.code
org 0100h

start:
        jmp init
label signature byte
        db 'interrupt$log'
label endsig byte
label myint byte
        db 0
label buffer_size word
        dw 01000h
label bufptr dword
        farptr <?,0>
label oldvec dword
        farptr ?

intvec:
                       ; bp+6  flags 1e
                       ; bp+4  seg   1c
                       ; bp+2  ofs   1a
        push  bp       ; bp+0        18
        mov   bp,sp
        push  bp       ; bp-2        16
        push  ds       ; bp-4        14
        push  es       ; bp-6        12
        push  ss       ; bp-8        10
        push  ax       ; bp-a         e
        push  bx       ; bp-c         c
        push  cx       ; bp-e         a
        push  dx       ; bp-10        8
        push  si       ; bp-12        6
        push  di       ; bp-14        4
        mov   ax,040h
        mov   ds,ax
        push  ds:[06eh] ; bp-16       2
        push  ds:[06ch] ; bp-18       0

; move stack frame to buffer
        mov   ax,ss
        mov   ds,ax
        les   di,cs:bufptr
        cmp   di, cs:buffer_size      ; do nothing if buffer full
        jae   exitint
        lea   si,[bp-18h]
        mov   cx,0010h                ; size of logged record
        cld
        repz  movsw
        mov   cs:bufptr.o,di          ; new buffer pointer
exitint:
        pop di
        pop di
        pop di
        pop si
        pop dx
        pop cx
        pop bx
        pop ax
        pop ss
        pop es
        pop ds
        pop bp
        pop bp
        jmp  cs:oldvec

; not used any more, otherwise can't locate  logint via int vector
; may I should try locating it through a MCB scan
restoreint:
        mov ax,cs
        mov ds,ax
        xor ax,ax
        mov es,ax
        mov al, cs:myint
        shl ax,1
        shl ax,1
        mov di,ax
        mov si, offset oldvec
        movsw
        movsw
        jmp exitint

init:
        mov ax,cs
        mov dx, offset init
        mov cl,4
        shr dx,cl
        inc dx
        add ax,dx
        mov bufptr.s,ax
        mov bl,byte ptr ds:[80h]
        xor bh,bh
        mov byte ptr [bx+81h],0    ; replace CR with terminating 0
        mov si,081h
        cld

option_loop:
        call  skip_spaces
        mov   al,[si]
        cmp al,'-'
        jne  getintno
        inc  si
        lodsb
        cmp al,'u'
        jne  opt1
        jmp  uninstall
; no other options jet
opt1:
        cmp al,'b'
        jne  opt2
        jmp  getbuffer
opt2:
        jmp option_loop
getbuffer:
        call skip_spaces
        call getword
        mov  buffer_size,ax
        jmp  option_loop

getintno:
        call gethex
        mov  myint,al
        mov ah,35h
        mov al,myint
        int 21h
        mov oldvec.o,bx       ; save original interrupt vector
        mov oldvec.s,es
        mov ah,25h
        mov al,myint
        mov dx,cs
        mov ds,dx
        mov dx, offset intvec
        int 21h             ; set new interrupt
        mov dx, buffer_size
        add dx, offset init
        mov cx,4
        shr dx,cl
        inc dx
        mov ax,03100h      ; terminate & stay resident
        int 21h
getword:
        call gethex
        mov  ah,al
        push ax
        mov  bp,sp
        call gethex
        mov  [bp],al
        pop  ax
        ret

gethex:
        lodsb
        call atoh
        mov ah,al
        lodsb
        call atoh
        aad  010h
        ret

atoh:
        cmp al,'9'
        jle atohend
        add al,9
atohend:
        and al,0fh
        ret

uninstall:
        call skip_spaces
        call gethex           ; get vector to unload
        mov myint,al
        mov Ah,35h
        mov al,myint
        int 21h
        mov oldvec.o,bx
        mov oldvec.s,es         ; save vector

        call check_sig        ; checks for signature in ES:
        jne   not_installed

        push ds
        lds dx,es:oldvec
        mov ah,25h              ; restore original vector
        mov al,cs:myint
        int 21h
        pop  ds

; write buffer to file
        call skip_spaces       ; skip to file name
        mov dx,si
        xor cx,cx              ; attributes
        mov ax,03c00h          ; create file
        int 21h
        mov bx,ax
        lds cx,es:bufptr       ; cx=length, ds:dx=buffer
        xor dx,dx
        mov ax,04000h
        int 21h                ; write file
        mov ax,03e00h          ; close file
        int 21h

; release memory
        mov ax,04900h      ; free psp es:0000
        int 21h
        mov ax,es:[02ch]
        mov es,ax
        mov ax,04900h      ; free environment
        int 21h
        mov ax,4c00h
        int 21h

not_installed:
        mov dx,offset errormsg
        mov ax,0900h
        int 21h
        mov ax,04c01h
        int 21h

; skips spaces in string pointed to by  ds:si
skip_spaces:
        push di
        push es
        push ax
        push cx
        push ds
        pop  es
        mov di,si
        mov al,' '
        mov cx,080h
        repe  scasb                ; skip spaces
        dec  di
        mov  si,di
        pop cx
        pop ax
        pop es
        pop di
        ret

; checks for signature in segment ES:
check_sig:
        push si
        push di
        push cx
        mov  di,offset signature
        mov  si,di
        mov  cx,offset endsig-signature
        cld
        repe cmpsb              ; check for signature
        pop  cx
        pop  di
        pop  si
        ret
errormsg:
        db 'logint was not found!', 0dh, 0ah, '$'
end start
