char *printTypes="RDxXoObBdDsS%";
#define PRT_OTHER      -1
#define PRT_REGISTER    0
#define PRT_HEXDUMP     1
#define PRT_HEX_lh      2
#define PRT_HEX_hl      3
#define PRT_OCTAL_lh    4
#define PRT_OCTAL_hl    5
#define PRT_BINARY_lh   6
#define PRT_BINARY_hl   7
#define PRT_DECIMAL_lh  8
#define PRT_DECIMAL_hl  9
#define PRT_STRING     10
#define PRT_LSTRING    11
#define PRT_PERCENT    12


PrintValue(long l, int base)
{
    long p;
    int i;
    for ( i=0 ; i < 3 ; i++  ){
        p=l%base;
        l/=base;
        putchar(p > 9 ? (p-10) + 'A': p + '0');
    }
}

byte *PrintHL(byte *buf, int len, int base)
{
  long l=0;
  int  shift=3;
  if (len>4) len=4;
  while (len--)
    l|=(*buf++)<<(shift-- *8);
  PrintValue(l, base);
  return buf;
}

PrintLH(byte *buf, int len, int base)
{
  long l=0;
  int  shift=0;
  if (len>4) len=4;
  while (len--)
    l|=(*buf++)<<(shift++ *8);
  PrintValue(l, base);
  return buf;
}

byte *PrintString(byte *buf, int len, int term)
{
  while (len-- && (term<0 || *buf!=term))
    putchar(*buf++);

  if (len)
    return buf+1;    // incl terminator
  else
    return buf;
}

byte *PrintPasString(byte *buf)
{
  int len;
  len=*buf++;
  return PrintString(buf, len, -1);
}

// gets paramters from <str>, separated by ';', terminated by '.'
// returns # params found
int GetParams(char *str, int *paramList)
{
  char *p=NULL;
  int  n=0;
  do {
    if (p) str=p+1;
    *paramList++=strtol(str, &p, 0);
    n++;
  } while (*p==';' && p>str);
  if (*p!='.')
    return 0;
  else
    return n;
}

// process string with format characters
//   %Rxx or %Rexx  (xx=registername)
//   %Dm;n.          print n bytes data from data offset m (dump format)
//   %xm;n.          print n bytes hex from data offset m (intel byte order)
//   %Xm;n.          print n bytes hex from data offset m (moterola byte order)
//      (o=oct, b=bin, d=dec)
//   %sm;t;n.        print <t> terminated string from offset m, max n bytes
//   %Sm.            print length preceeded string from offset m
//   m == x[,y]  ,y == offset x after y-th string
//  [;n] is optional
//
int rprintf(char *fmt, struct int_rec *rec)
{
  char *p;
  int  fmtType;
  int  i;
  char regName[5];
  char *regVal;
  byte *data;

  int  param[5];

  data=(byte *)&rec[1];
  offset=0;
  endOffset=rec->reclen-sizeof(*rec);

  while (fmt[i])
  {
    if (fmt[i]=='%')
    {
      p=strchr(printTypes, fmt[i+1]);
      if (p==NULL)
        fmtType=PRT_OTHER;
      else
        fmtType=p-printTypes;
      nParams=GetParams(&fmt[i+2], param);
      length=rec->reclen-sizeof(*rec)-offset;
      switch(fmtType)
      {
        case PRT_REGISTER  :
            regname[0]=fmt[i+2];
            regname[1]=fmt[i+3];
            regname[2]=0;
            regVal=LookupRegister(regName, rec);
            if (regVal)
              fputs(regVal, stdout);
            else
              printf("\ninvalid registername\n");
            break;
        case PRT_HEXDUMP   :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            dump(data+offset, length);
            offset+=length;
            break;
        case PRT_HEX_lh    :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintLH(data+offset, length, 16);
            break;
        case PRT_HEX_hl    :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintHL(data+offset, length, 16);
            break;
        case PRT_OCTAL_lh  :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintLH(data+offset, length, 8);
            break;
        case PRT_OCTAL_hl  :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintHL(data+offset, length, 8);
            break;
        case PRT_BINARY_lh :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintLH(data+offset, length, 2);
            break;
        case PRT_BINARY_hl :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintHL(data+offset, length, 2);
            break;
        case PRT_DECIMAL_lh:
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintLH(data+offset, length, 10);
            break;
        case PRT_DECIMAL_hl:
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            offset=PrintHL(data+offset, length, 10);
            break;
        case PRT_STRING    :
            terminator=-1;
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            if (nParams>1 && param[1]>0 && param[1]<length) length=param[1];
            if (nParams>2 && param[2]>=0) terminator=param[2];
            offset=PrintString(data+offset, length, terminator);
            break;
        case PRT_LSTRING   :
            if (nParams>0 && param[0]>0 && param[0]<endOffset) offset=param[0];
            offset=PrintPasString(data+offset);
            break;
        case PRT_PERCENT   :
            putchar('%');
          break;
        case PRT_OTHER     :
        default:
          printf("\nError in format string:
          // unknown print type
      }
      p=strchr(&fmt[i],'.');
      if (p==NULL)
        printf("\nError in format %s\n", fmt);
      else
        i+= p - &fmt[i]+1;
    }
  }
  return 0;
}
