;
;  catchx is an unfinished attempt to universaly determine which interrupt
;  occurred.
;           
; Catch interrupt V2.0  - 10 jan 1994
;                  by   - Willem Jan Hengeveld
;               email   - willem@dutikos.twi.tudelft.nl
;               phone   - +31-15-617267
;
;  a program to log all calls to interrupts specified on the command line
;
;  Usage: catch [s]i <intno> ...
;            S = only store 'specials'
;         catch u <filename>
;         a maximum of 10 interrupts can be monitored at one time
;
; Plans:
;   - convert to bigger memory model
;   - think of a better way to implement specialactions
;   - allow for return values to be stored
;   - I am not sure if time calculation is correct -> check
;   - allow interrupt list to be edited when installed
;   - no check is made to see if catch is already installed
;   - put version number in signature to check for correct version
;   - add more special actions
;   - construct all the 'catchX' routines with 1 macro, or maybe add them
;     dynamicly at runtime
;   - option to display current status
;   - clean up commandline processor
;
;   - add exclusion list: certain CS/IP/AX values can be excluded from
;     logging
;   - reduce storage: 1 word specifies which registers have to be stored
;     bits0-7=ax cx dx bx sp bp si di   (modrm order)
;     bits8-d=es cs ss ds ip sw
;   - make code reentrant so hardware interrupts will be logged properly
;   - change order of registers to match modrm order (this might come in
;     handy when processing instructions
;   - single step option (for instance when scrolllock is on)
;   - start/stop logging with some hotkey
;   - stay on top option: check if address in int table matches my entry point
;     if not try to find out who changed it.
;   - trap int2125 to allways stay on top
;   - flexible changing of MAXINTS
;   - may be it is possible to find out what caused int:
;     - check instruction at CS:[IP-2] == INT xx
;     - check PIC to see if hw int occurred
;          out 20,0a  in al,20   -> read isr
;          out 20,0b  in al,20   -> read irr
;

.286
biosseg segment at 040h
org 0
  rs232portlist     dw 4 dup(?)           ; 0000
  lptaddr           dw 4 dup(?)           ; 0008
  hardwareinst      dw ?                  ; 0010
  post_status       db ?                  ; 0012
  basememsize       dw ?                  ; 0013
  scratch_1         dw ?                  ; 0015
  keystatus         dw ?                  ; 0017
  scratch_2         db ?                  ; 0019
  nextkeyptr        dw ?                  ; 001a
  nextfreeptr       dw ?                  ; 001c
  keybuffer         dw 010h dup(?)        ; 001e
  driverecalstat    db ?                  ; 003e
  drivemotorstat    db ?                  ; 003f
  drivemotoroffcnt  db ?                  ; 0040
  drivelastresult   db ?                  ; 0041
  drivecmd          db 7 dup(?)           ; 0042
  currvideomode     db ?                  ; 0049
  videocols         dw ?                  ; 004a
  videopagesize     dw ?                  ; 004c
  curpageofs        dw ?                  ; 004e
  cursorpos         dw 08h dup(?)         ; 0050
  cursortype        dw ?                  ; 0060
  currentpage       db ?                  ; 0062
  crtportaddr       dw ?                  ; 0063
  crtmodeselect     db ?                  ; 0065
  crtcgapallette    db ?                  ; 0066
  postresetreentry  dd ?                  ; 0067
  postlastirq       db ?                  ; 006b
  biosTimerL        dw ?                  ; 006c
  biosTimerH        dw ?                  ; 006e
  biostimeroverflow db ?                  ; 0070
  ctrlbreakflag     db ?                  ; 0071
biosseg ends

mcbseg segment at ?
  mcbflag           db ?
  mcbowner          dw ?
  mcblen            dw ?
mcbseg ends

RECSIZE    equ 024h      ; # words in default record
MAXRECSIZE equ 0e0h
MAXINTS    equ 10        ; # int buffers
MAXRECDATA equ MAXRECSIZE - RECSIZE   ; # bytes that may be appended to a record

code  segment para 'CODE'
  assume  cs:code, ds:code, es:code, ss:code

psp_int20       db 2 dup(?)
psp_nextseg     dw ?
psp_unused1     db ?
psp_cpmcall5    db ?
psp_seg1len     dw ?
psp_cpmcall5_2  db 2 dup(?)
psp_int22vec    dd ?
psp_int23vec    dd ?
psp_int24vec    dd ?
psp_parentseg   dw ?
psp_jobfiletab  db 20 dup(?)
psp_envseg      dw ?
psp_savsssp     dd ?           ; on last int21 call
psp_njft        dw ?
psp_jftptr      dd ?
psp_prevpsp     dd ?
psp_unused2     db 4 dup(?)
psp_dosver      dw ?
psp_unused3     db 26 dup(?)
psp_fcb1        db 16 dup(?)
psp_fcb2        db 16 dup(?)
psp_unused4     db 4 dup(?)
psp_cmdlen      db ?
psp_cmdline     db 126 dup(?)

org 0100h

start:
     jmp  _init

label signature byte
        db 'interrupt$catch'
label endsig byte

even      ; for speed
_intcounter label word
     dw   MAXINTS dup(0)
_nentries label word
     dw   0
_maxbuf  label word
     dw   0
_lastentry label dword
     dw   0
     dw   0
_buffer  label dword
     dw   0
     dw   0
_oldint label dword
     dd   0
_oldinttab  label dword
     dd   MAXINTS dup(0)
_catchtab label word      ; this should contain a list of _catch0 .. _catchN
     dw   _catch0, _catch1, _catch2, _catch3
     dw   _catch4, _catch5, _catch6, _catch7
     dw   _catch8, _catch9
_intnrs  label byte
     db   MAXINTS dup(0)
_curint label byte         ; _intnrs [ _curintno ]
     db   0
_curintno label byte       ; index into _intnrs tab of _curint
     db   0
_nints label word          ; # entries in _intnrs table
     dw   0
_video_seg label word
     dw   0b000h
_specialonly label byte    ; commandline flag
     db   0
assume ds:nothing, es:nothing, ss:nothing
;*************************************************************
; convert ax -> ascii string and print at position di on topline of screen
; input: AX, DI
; return:  BX=0a, CX=0, AX=?? DX=0, DI=??  ES=video
;
_display proc near
     mov bx,_video_seg
     mov es,bx
     mov bx,0ah
     mov cx,5
     std
convloop:
     xor dx,dx
     div  bx
     xchg dx,ax
     mov  ah,070h
     add  al,'0'
     stosw
     xchg dx,ax
     test ax,ax
     je   fillup
     loop convloop
     ret
 fillup:
     mov ax,7020h
     rep stosw
     ret
_display endp

;*************************************************************
; read microsecond accuracy timer
; input: -
; output: DX:AX = time
;
_gettime proc near
     push bp
     push ds
     mov  bp,sp
     mov  ax,40h
     mov  ds,ax
assume ds:biosseg
     push biosTimerH  ; bp-2
     push biosTimerL  ; bp-4
     xor al,al
     out  43h,al
     in   al,40h
     mov ah,al
     in   al,40h
     push ax          ; bp-6
     push biosTimerH  ; bp-8
     push biosTimerL  ; bp-a
     xor al,al
     out  43h,al
     in   al,40h
     mov ah,al
     in   al,40h
     push ax          ; bp-c
     mov  ax,[bp-00h]
     xor  ax,[bp-0ah]
     jnz  tmr_unequal
     mov  ax,[bp-02h]
     xor  ax,[bp-08h]
     jnz  tmr_unequal
     mov  dx,[bp-04h]        ; take first
     mov  ax,[bp-06h]
     jmp  end_tmr
tmr_unequal:
     mov  dx,[bp-0ah]        ; take second
     mov  ax,[bp-0ch]
end_tmr:
     xchg ah,al
     neg  ax
     mov  sp,bp
     pop  ds
     pop  bp
     ret
_gettime endp

;*************************************************************
; initializes timer chip for usec timing
; input: -
; output: AL=0
;
_inittimer proc near
     mov al,34h
     out 43h,al
     mov al,0
     out 40h,al
     out 40h,al
     ret
_inittimer endp

;*************************************************************
; restore timer chip to normal mode
; input: -
; output: AL=0
;
_restoretimer proc near
     mov al,36h
     out 43h,al
     mov al,0
     out 40h,al
     out 40h,al
     ret
_restoretimer endp

;*************************************************************
; table with actions for special values
;    0 - normal
;    1 - copy ascii0d from dssi
;    2 - copy asciiz string from dsdx
;    3 - append to previous
;    4 - (1), but only if AL==33
;    5 - copy Lstr from dssi, Lstr from dsdx [0d]
_action  label byte
      db  3, 1, 2, 2, 2, 2, 4, 5

;*************************************************************
; table with values for intno & AH which require special action
;
_special label word
      dw  02102h, 02129h, 0213ch, 0213dh, 0214bh, 0214eh, 02f11h, 02faeh

;*************************************************************
;  checks if AX contains special int function  CX and DI are destroyed
;  input: AX
;  output: AX=??, CX=??, DI=??
;
_checkspecial proc near
      mov cx, offset _special - offset _action
      mov di, 0
checkloop:
      cmp  _special[di], ax
      jz   foundspecial
      inc  di
      inc  di
      loop checkloop
return0:
      xor ax,ax
      ret
foundspecial:
      shr  di,1
      mov  al, _action[di]
      xor  ah, ah
      ret
_checkspecial endp

;*************************************************************
; movestring - copies string ds:si -> es:di, adjust length of record
; maximum length=CX,  terminating char=AL  or 0
; input: DS:SI, AL, CX
; output: CX,  SI=si+CX, DI=lastentry, AH=al
movestring proc near
     cld
     mov  ah,al
     les  di, _buffer
movestrloop:
     lodsb
     stosb
     or   al,al
     jz   end_of_string
     cmp  al,ah
     jz   end_of_string
     loop   movestrloop
end_of_string:
     mov  cx,di
     sub  cx, word ptr _buffer
     mov  word ptr _buffer, di
     mov di,word ptr _lastentry
     add  es:[di], cl             ; adjust record length
     ret
movestring endp
;*************************************************************
;  gets IRR & ISR from PIC #1 in AX
;  gets IRR & ISR from PIC #2 in BX
getisrirr proc near
     mov al,0bh       ; get irr
     out 0a0h,al
     in  al,0a0h
     mov ah,al
     mov al,0ah       ; get isr
     out 0a0h,al
     in  al,0a0h
     mov bx,ax
     mov al,0bh       ; get irr
     out 20h,al
     in  al,20h
     mov ah,al
     mov al,0ah       ; get isr
     out 20h,al
     in  al,20h
     ret
getisrirr endp
;*************************************************************
;  lookup current int in interrupt table
;  return C=0, DI-> intrec   if int found
;         C=1, DI=0          if int not found
lookupint proc near
     mov  di,offset intlist
     lodsb
     cmp  ah,al
     je   found
     add  di, sizeof(struc intrec)-1
     cmp  di, endofintlist
     jl   lookuploop
     stc
     xor di,di
     ret
found:
     clc
     dec di
     ret
lookupint endp

;*************************************************************
; store processor status in buffer, display counter, ...
;
_catchint proc near
assume ds:nothing, es:nothing, ss:nothing

                        ; bp-02  flags 22
                        ; bp-04  CS    20
                        ; bp-06  IP    1e
     push bp            ; bp-08  BP    1c
     mov  bp, word ptr _buffer
     cmp  bp, word ptr _maxbuf
     jb   process_int
     jmp  buf_full
process_int:
     mov  bp,sp
     add  bp,8
     push bp            ; bp-0a  SP   1a
     push ss            ; bp-0c  SS   18
     push di            ; bp-0e  DI   16
     push es            ; bp-10  ES   14
     push si            ; bp-12  SI   12
     push ds            ; bp-14  DS   10
     push dx            ; bp-16  DX   0e
     push cx            ; bp-18  CX   0c
     push bx            ; bp-1a  BX   0a
     push ax            ; bp-1c  AX   08
     mov  ax,cs
     mov  ds,ax
assume ds:code
     les  si,[bp-6]
     cmp  byte ptr es:[si-2],0cdh
     jne  ishw
     mov  ah,es:[si-1]
     jmp intlocated
ishw:

intlocated:
     call  lookupint
     call _gettime
     push dx            ; bp-1e  timH 06
     push ax            ; bp-20  timL 04
     mov  bl, _curintno
     xor  bh,bh
     shl  bx,1
     push _intcounter[bx]   ; bp-22  n    02
     mov  ah, _curint
     xor  al, al        ;    length of record
     push ax            ; bp-24  int  00
     mov  si,sp
     mov  byte ptr ss:[si], RECSIZE
     call getisrirr
     mov  [bp-1eh],ax
     mov  [bp-20h],bx
     mov ah,[bp-23h]     ; intno
     mov al,[bp-1bh]     ; AH
     call _checkspecial
     mov  ah,_specialonly
     cmp  ax,0100h
     je   skip_copy
     cmp  al,3
     jne  normal_copy
     jmp   append_to_prev
skip_copy:
     jmp   end_of_copy
; all regs are copied into buffer
;  AL has to have special action value
normal_copy:
     inc  _nentries
     les  di, dword ptr _buffer
     mov  word ptr _lastentry, di
     mov  word ptr _lastentry+2, es
     push ss
     pop  ds
     mov  cx,RECSIZE              ; move bytes because SP might be odd
     rep  movsb                   ; mov [sp=ffff] .. causes GPI
     mov  word ptr _buffer, di

; process further copy specifications

     cmp  al,0
     je   go_endoc
     cmp  al,1
     jne  checkal2

; append dssi [bp-14] string to buffer
     mov  si,[bp-12h]
     mov  ds,[bp-14h]
     mov  cx,MAXRECDATA
     mov  al,0dh
     call movestring
go_endoc:
     jmp  end_of_copy

checkal2:
     cmp  al,2
     jne  checkal4

; append dsdx string to buffer
     mov  ds,[bp-14h]
     mov  si,[bp-16h]
     mov  cx,MAXRECDATA
     mov  al,0
     call movestring
     jmp  end_of_copy

checkal4:
     cmp  al,4
     jne  checkal5

; append dssi string to buffer  if int2f1123
     cmp  byte ptr [bp-1ch],023h
     jne  end_of_copy
     mov  si,[bp-12h]
     mov  ds,[bp-14h]
     mov  cx,MAXRECDATA
     mov  al,0
     call movestring
     jmp  end_of_copy

checkal5:
     cmp  al,5
     jne  checkal5

; append dssi, dsbx string to buffer if int2fae
     mov  si,[bp-12h]    ; copy dssi string
     mov  ds,[bp-14h]
     mov  cl,ds:[si]
     xor  ch,ch
     mov  al,0
     call movestring

     mov  ds,[bp-14h]    ; copy dsbx string
     mov  si,[bp-1ah]
     inc  si             ; skip one byte
     mov  cl,ds:[si]
     xor  ch,ch
     mov  al,0dh
     call movestring
     jmp  end_of_copy

go_copy:
     mov  al,0           ; no further action
     jmp normal_copy

append_to_prev:
     mov ah,[bp-23h]     ; intno
     mov al,[bp-1bh]     ; AH
     les di, _lastentry
     cmp ah,es:[di+1]       ; check if lastentry was the same
     jne go_copy
     cmp al,es:[di+9]
     jne go_copy
     mov al,[bp-16h]     ; DL register
     mov cl,es:[di]
     mov ch,0
     cmp cx,MAXRECSIZE   ; maximum length in one record
     jge go_copy
     inc  byte ptr es:[di]
     les di, _buffer
     stosb
     mov  word ptr _buffer, di

end_of_copy:
     mov  bl, _curintno
     xor  bh,bh
     shl  bx,1
     inc  _intcounter[bx]
     add  sp,8
     mov  ax, _intcounter[bx]
     shl  bx,1
     mov  di,bx
     shl  bx,1      ; on top line 6 positions for each counter
     add  di,bx
     add  di,00eh
     call _display
     mov  ax, word ptr _buffer
     mov  di, 092h
     call _display
     mov  ax, _nentries
     mov  di, 09eh
     call _display
     pop  ax
     pop  bx
     pop  cx
     pop  dx
     pop  ds
     pop  si
     pop  es
     pop  di
     pop  ss
     pop  bp

buf_full:
     mov bp,sp
     mov bp,[bp+6]     ; restore flags (important)
     push bp
     popf
     pop bp
     jmp _oldint
_catchint endp

;*************************************************************
; install catch in memory
;
_tsr  proc near
     call _inittimer
     mov  _maxbuf,0ff00h
     mov  word ptr _buffer, offset endofresidentcode
     mov  word ptr _buffer+2, cs
     mov  dx, 01000h
     mov  ax,3100h
     int  21h
_tsr  endp
;  this is where the buffer starts
;
endofresidentcode:

;*************************************************************
; positions SI to next non space character
;
_skipcmdspaces proc near
     or   cx,cx
     je   ret_eolskip
     mov  al, ' '
     xchg  di,si
     push es
     push ds
     pop  es
     repz scasb
     xchg  di,si
     dec  si
     inc  cx
     pop  es
     clc
     ret
ret_eolskip:
     stc
     ret
_skipcmdspaces endp

;*************************************************************
; gets char pointed to by SI, if CX>0
;
_getcmdchar proc near
     or   cx, cx
     je   ret_eol
     lodsb
     dec cx
     clc
     ret

ret_eol:
     stc
     ret
_getcmdchar endp

_gethexbyte proc near
        call _skipcmdspaces
        jb   gh_eol
        lodsb
        dec   cx
        call atoh
        mov ah,al
        lodsb
        dec   cx
        call atoh
        aad  010h
        clc
        ret
atoh:
        cmp al,'9'
        jle atohend
        add al,9
atohend:
        and al,0fh
        ret

gh_eol:
        stc
        ret
_gethexbyte endp

;*************************************************************
; install handler for interrupt in AL
;  - buffer and maxbuf are both 0 while installing ->
;    the catch routine is called but skipped due to full buffer
_install_handler proc near
     mov  di, _nints
     cmp  di,MAXINTS
     ja   cantinstall
     mov  _intnrs[di], al
     mov  ah,035h
     int  21h
     shl  di,1
     shl  di,1
     mov  word ptr _oldinttab[di], bx
     mov  word ptr _oldinttab[di+2], es
     mov  ah, 025h
     shr  di,1
     mov  dx, _catchtab[di]
     int  021h
     inc  _nints
     clc
     ret
cantinstall:
     stc
     ret
_install_handler endp

; checks for signature in segment ES:
check_sig proc near
        push si
        push di
        push cx
        mov  di,offset signature
        mov  si,di
        mov  cx,offset endsig-signature
        cld
        repe cmpsb              ; check for signature
        pop  cx
        pop  di
        pop  si
        ret
check_sig endp
;*************************************************************
; locate previously loaded catch by scanning through MCB's
;
_locate_catch proc near
     mov ah,052h
     int 21h
     mov bx,es:[bx-2]             ; get 1st mcb from dosvars

loc_loop:
     mov es,bx
assume es:mcbseg
     cmp  mcbflag,'Z'
     je   signotfound
     add bx,word ptr es:mcblen
     inc bx
     mov ax,word ptr es:mcbowner  ; segment of owner of MCB
     mov dx,cs
     cmp ax,dx                    ; don't find myself
     je  loc_loop
     mov es,ax
assume es:nothing

     call check_sig
     jne  loc_loop

     clc
     ret
signotfound:
     stc
     ret

_locate_catch endp

calc_videoseg proc near
     mov ah, 0fh
     int 10h
     cmp al, 7
     je  nochange
     mov _video_seg,0b800h
nochange:
     ret
calc_videoseg endp


;*************************************************************
; startup
;
_init proc near
     call  calc_videoseg
     cld
     mov  si, offset psp_cmdline
     mov  cl, psp_cmdlen
     mov  ch, 0
     mov  di,si
     add  di,cx
     xor  al,al
     stosb               ; put 0 at end of command line
next_cmd:
     call _skipcmdspaces
     jc   do_install
     call _getcmdchar
     and  al,0dfh        ; upper case
     cmp  al,'U'
     je   do_uninstall
do_install:
     cmp  al,'I'
     je   chkinstall
     cmp  al,'S'
     je   enable_spec
     jmp  do_help
enable_spec:
     mov  _specialonly,1
     jmp  next_cmd
chkinstall:
     call _locate_catch      ; ES now points to catch PSP
     mov  dx,offset txt_found
     jb   installloop
     jmp  exit_err
installloop:
     call _gethexbyte
     jb   go_tsr             ; no more digits -> tsr
     call _install_handler   ; installs handler for INT [num]
     jnb  installloop
go_tsr:
     jmp  _tsr
do_uninstall:
     call _locate_catch      ; ES now points to catch PSP
     mov  dx,offset txt_notfound
     jb    exit_err
     mov  di,es:_nints
     push  es
     push  ds
rmvloop:
     or   di,di
     jz   endrmv
     dec  di
     mov  al,es:_intnrs[di]
     push ax
     mov  dx,es
     mov  ah, 035h
     int  21h
; make sure vector still pointing to catch
     mov  ax,es
     mov  es,dx
     shl  di,1
     cmp  ax,dx
     jne  vecchanged
     cmp  es:_catchtab[di],bx
     jne  vecchanged
     shl  di,1
     mov  dx,word ptr es:_oldinttab[di]
     mov  ds,word ptr es:_oldinttab[di+2]
     pop  ax
     mov  ah,025h
     int  21h
     shr  di,2
     jmp rmvloop
vecchanged:
; here we should try to locate the vector in memory
; and change it. --- risky
     jmp rmvloop

endrmv:
;  now all ints have been restored, we can write buffer to a file
        pop   ds
        pop   es
        call _skipcmdspaces    ; skip to file name
        jb   releasemem
        mov dx,si
        xor cx,cx              ; attributes
        mov ax,03c00h          ; create file
        int 21h
        mov bx,ax
        mov  dx, offset endofresidentcode
        lds cx,es:_buffer      ; cx=length, ds:dx=buffer
        sub  cx,dx
        mov ax,04000h
        int 21h                ; write file
        mov ax,03e00h          ; close file
        int 21h

releasemem:
        mov ax,04900h      ; free psp es:0000
        int 21h
        mov es,es:psp_envseg
        mov ax,04900h      ; free environment
        int 21h
        mov ax,4c00h
        int 21h

exit_err:
       mov   ah,9
       int   21h
       mov   ax,4c01h
       int   21h

do_help:
        mov  ah,9
        mov  dx,offset txt_help
        int  21h
        mov  ax,4c00h
        int 21h
_init endp
txt_notfound label byte
        db  'Catch not found.', 0dh, 0ah, '$'
txt_found label byte
        db  'Catch already running.', 0dh, 0ah, '$'
txt_help label byte
        db  'CATCH V2.0 - 1994 willem jan hengeveld', 0dh, 0ah
        db  'Usage: catch [S] I intno ...', 0dh, 0ah
        db  '       S - only store specials', 0dh, 0ah
        db  '   or  catch U filename', 0dh, 0ah, '$'
code ends

;        public _buffer:dword
;        public _maxints:word
;        public _intcounter:word
;        public _old21:dword
;        public _catch21
;        public _inittimer
;        public _restoretimer
        end start
