#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <io.h>

#define MAXRECSIZE 0x100

typedef unsigned int word;
typedef unsigned char byte;
struct int_rec {
  byte reclen;
  byte intno;
  int n;
  long time;
  word AX;
  word BX;
  word CX;
  word DX;
  word DS;
  word SI;
  word ES;
  word DI;
  word SS;
  word SP;
  word BP;
  word IP;
  word CS;
  word flags;
};

struct name_tab {
  int num;
  char *name;
  int display;
} intnametab[]={
  0x2102, "stdout",        0, /* string */
  0x2104, "stdaux",        0, /* string */
  0x2105, "prn",           0, /* string */
  0x2106, "con",           0, /* string */
  0x2109, "stdout",        0, /* string */
  0x210e, "setdrv",        4, /* DL */
  0x2119, "getdrv",        0, /* - */
  0x2129, "parse",         0, /* string */
  0x2139, "mkdir",         0, /* string */
  0x213a, "rmdir",         0, /* string */
  0x213b, "chdir",         0, /* string */
  0x213c, "creat",         1, /* string, CX */
  0x213d, "open",          2, /* string, AL */
  0x213e, "close",         5, /* BX */
  0x213f, "read",          3, /* BX, CX, DSDX */
  0x2140, "write",         3, /* BX, CX, DSDX */
  0x2141, "unlink",        0, /* string */
  0x2142, "lseek",         7, /* BX, CXDX, AL */
  0x2143, "fattr",         8, /* string, AL, CX */
  0x2147, "getcwd",        4, /* DL */
  0x214b, "exec",          2, /* string, AL */
  0x214e, "findfirst",     1, /* string, CX */
  0x214f, "findnext",      0, /* string */
  0x2156, "rename",        0, /* string */
  0x215a, "creattmp",      1, /* string, CX */
  0x215b, "creatnew",      1, /* string, CX */
  0x215c, "lock",          6, /* BX, AL, CXDX, SIDI */
  0x216c, "extended open/create",  9, /* string, BX, CX, DX */
  -1, NULL, -1
};

char *getstring(struct name_tab *tab, int n, int *display)
{
  while (tab->name)
  {
    if (tab->num==n)
    {
      *display=tab->display;
      return tab->name;
    }
    tab++;
  }
  return NULL;
}

void printrec(struct int_rec *p)
{
  char *data=(char *)p+sizeof(*p);
  int i;
  char *name;
  int  namehash;
  int outside;  /* flag to tell whether we are inside or outside quotes */
  int disp;     /* display format */

  printf("%08lx ", p->time);
  printf("%04x:%04x ", p->CS, p->IP);

  namehash=((p->AX>>8)&0xff)|(p->intno<<8);
  name=getstring(intnametab, namehash, &disp);
  if (name==NULL) name="?";
  printf("%s ", name);  outside=1;
  switch (disp)
  {
    case 0:  /* just print string */
      break;
    case 1:  /* print CX */
      printf("CX=%04x ", p->CX);
      break;
    case 2:  /* print AL */
      printf("AL=%02x ", p->AX&0xff);
      break;
    case 3:  /* print BX, CX, DSDX */
      printf("BX=%04x, CX=%04x, DSDX=%04x:%04x ", p->BX, p->CX, p->DS, p->DX);
      break;
    case 4:  /* print DL */
      printf("DL=%02x ", p->DX&0xff);
      break;
    case 5: /* BX */
      printf("BX=%04x ", p->BX);
      break;
    case 6: /* BX, AL, CXDX, SIDI */
      printf("BX=%04x AL=%02x CXDX=%04x%04x SIDI=%04x%04x ",
        p->BX, p->AX&0xff, p->CX, p->DX, p->SI, p->DI);
      break;
    case 7: /* BX, CXDX, AL */
      printf("BX=%04x CXDX=%04x%04x AL=%02x ",
        p->BX, p->CX, p->DX, p->AX&0xff );
      break;
    case 8: /* string, AL, CX */
      printf("AL=%02x CX=%04x ",
        p->AX&0xff, p->CX );
      break;
    case 9: /* string, BX, CX, DX */
      printf("BX=%04x CX=%04x DX=%04x ",
        p->BX, p->CX, p->DX);
      break;
    default:
      printf("Unknown fmt ");
  }
  if (p->reclen>sizeof(*p))
  {
    for (i=sizeof(*p) ; i<p->reclen ; i++)
      if (*data<' ')
      {
        if (!outside)
        {
          printf("\" ");
          outside=1;
        }
        printf("%02x ",0xff & (*data++));
      }
      else
      {
        if (outside)
        {
          printf("  \"");
          outside=0;
        }
        putchar(*data++);
      }
    if (!outside) printf("\"");
  }
  putchar('\n');
}

int usage(char *s, ...)
{
  va_list ap;
  va_start(ap,s);
  vprintf(s,ap);
  va_end(ap);
  printf("Usage: lstfile [-h] catchfile\n");
  return 1;
}

int main(int argc, char **argv)
{
  char *dmpname=NULL;
  int f;
  char data[MAXRECSIZE];
  struct int_rec *buf;
  long t0;

  while (argc>1)
  {
    if (argv[1][0]=='-')
    switch(argv[1][1])
    {
      case 'h': return usage("Help screen\n");
      default:  return usage("Invalid option: %s\n", argv[1]);
    }
    else if (dmpname==NULL)
      dmpname=argv[1];
    else
      return usage("Too many parameters: %s\n", argv[1]);

    argc--; argv++;
  }
  if (dmpname==NULL)
    return usage("You have to specify catch dumpfile\n");

  f=open(dmpname, O_RDONLY|O_BINARY);
  if (f<0) { perror(dmpname); return 1; }

  printf("                 time       CS   IP   data\n");
  t0=0;
  while (!eof(f))
  {
    read(f,data, sizeof(*buf));
    buf=(struct int_rec *)data;
    if (t0)
      printf("%7lx ", (buf->time)-t0);
    else
      printf("        ");
    t0=buf->time;
    if (buf->reclen>sizeof(*buf))
      read(f, data+sizeof(*buf), buf->reclen-sizeof(*buf));
    printrec(buf);
  }
  return close(f);
}
