#include <io.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long  LONG;

struct pack {
  WORD timer;
  WORD bios;
  WORD irsr;
};
//
//  2 first
//      sti/cli
//  1 middle
//      sti/nop/cli
//  0 last
//
struct frame {
  struct pack d[3];
  WORD flags;
};

#define TIMER_WRAP  1
#define IRRISR      2
#define BIOS_CHG    4
#define BIOS_INC    8

void main(int argc, char **argv)
{
  int f;
  struct frame buf;
  WORD prevChip=0;
  WORD prevBios=0;
  int  flag;
  int  i;
  long t, t0;
  if (argc<=1) return;

  f=open(argv[1],O_RDONLY|O_BINARY);
  if (f<0) return;
  perror("open");
  while (read(f,&buf,sizeof(buf))==sizeof(buf))
  {
    t=buf.d[0].bios;
    t=(t<<16)-buf.d[0].timer;
    printf("%5ld : ", t-t0);
    t0=t;
    for (i=2 ; i>=0 ; i--)
      printf("%04x %04x:%04x - ",
        buf.d[i].irsr, buf.d[i].bios, buf.d[i].timer);
    printf("%04x", buf.flags);
    flag=0;
    if (buf.d[0].bios != buf.d[2].bios || buf.d[1].bios != buf.d[2].bios)
      flag|=BIOS_CHG;
    if (buf.d[0].irsr || buf.d[1].irsr || buf.d[2].irsr)
      flag|=IRRISR;
    if (prevBios!=buf.d[2].bios)
      flag|=BIOS_INC;
    if (prevChip<buf.d[2].timer)
      flag|=TIMER_WRAP;

    if (flag)
      printf(" *%x\n", flag);
    else
      putchar('\n');
    prevChip=buf.d[2].timer;
    prevBios=buf.d[2].bios;
  }
  perror("read");
  close(f);
  perror("close");
}
/*
bios = bios timer tick teller (0040:006c)
pit  = gelatchte waarde gelezen van de timer chip
isr  = inservice register van de interrupt controller

            sti/cli        sti/nop/cli
isr  bios pit    isr  bios pit    isr  bios pit    flags
     0329:012b -      0329:011a -      0329:0109 - 7046
 1   0329:fff2 -  1   0329:ffe2 -      032a:ff84 - 7046 *7
     032a:fe63 -      032a:fe53 -      032a:fe42 - 7046 *8

     032c:0139 -      032c:0129 -      032c:0118 - 7046
     032c:0001 -  1   032c:fff0 -      032d:ff93 - 7046 *6
     032d:fe73 -      032d:fe62 -      032d:fe51 - 7046 *9

     0330:012f -      0330:011e -      0330:010d - 7046
 1   0330:fff6 -  1   0330:ffe6 -      0331:ff89 - 7046 *7
     0331:fe68 -      0331:fe57 -      0331:fe46 - 7046 *8

     0335:0142 -      0335:0132 -      0335:0121 - 7046
     0335:000b -      0335:fffa -      0336:ff9d - 7046 *4
     0336:fe7f -      0336:fe6e -      0336:fe5d - 7046 *9
*/
