#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <io.h>

#define MAXRECSIZE 0x100

typedef unsigned int word;
typedef unsigned char byte;
struct int_rec {
  byte reclen;
  byte intno;
  byte irr;
  byte isr;
  long time;
  word AX;
  word BX;
  word CX;
  word DX;
  word DS;
  word SI;
  word ES;
  word DI;
  word SS;
  word SP;
  word BP;
  word IP;
  word CS;
  word flags;
};

void printrec(struct int_rec *p)
{
  char *data=(char *)p+sizeof(*p);
  int i;
  char *name;
  int  namehash;
  int outside;  /* flag to tell whether we are inside or outside quotes */
  int disp;     /* display format */

  printf("%08lx ", p->time);
  printf("%02x %02x ", p->irr, p->isr);
  printf("%04x:%04x ", p->CS, p->IP);
  printf("%04x\n", p->flags);
}

int usage(char *s, ...)
{
  va_list ap;
  va_start(ap,s);
  vprintf(s,ap);
  va_end(ap);
  printf("Usage: lstfile [-h] catchfile\n");
  return 1;
}

int main(int argc, char **argv)
{
  char *dmpname=NULL;
  int f;
  char data[MAXRECSIZE];
  struct int_rec *buf;
  long t0;

  while (argc>1)
  {
    if (argv[1][0]=='-')
    switch(argv[1][1])
    {
      case 'h': return usage("Help screen\n");
      default:  return usage("Invalid option: %s\n", argv[1]);
    }
    else if (dmpname==NULL)
      dmpname=argv[1];
    else
      return usage("Too many parameters: %s\n", argv[1]);

    argc--; argv++;
  }
  if (dmpname==NULL)
    return usage("You have to specify catch dumpfile\n");

  f=open(dmpname, O_RDONLY|O_BINARY);
  if (f<0) { perror(dmpname); return 1; }

  printf("    diff  time   irr isr  CS:IP   flags\n");
  t0=0;
  while (!eof(f))
  {
    read(f,data, sizeof(*buf));
    buf=(struct int_rec *)data;
    if (t0)
      printf("%8lx ", (buf->time)-t0);
    else
      printf("         ");
    t0=buf->time;
    if (buf->reclen>sizeof(*buf))
      read(f, data+sizeof(*buf), buf->reclen-sizeof(*buf));
    printrec(buf);
  }
  return close(f);
}
