#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include "lst.h"

#define MAXRECSIZE 0x100

//  0 : string                     "%D"
//  1 : string + CX                "cx=%Rcx  %D"
//  2 : string + AL                "al=%Ral  %D"
//  3 : BX + CX + DSDX             "bx=%Rbx cx=%Rcx dsdx=%Rds:%Rdx"
//  4 : DL                         "dl=%Rdl"
//  5 : BX                         "bx=%Rbx"
//  6 : BX + AL + CXDX + SIDI      "bx=%Rbx al=%Ral cxdx=%Rcx%Rdx sidi=%Rsi%Rdi"
//  7 : BX + CXDX + AL             "bx=%Rbx cxdx=%Rcx%Rdx al=%Ral"
//  8 : string + AL + CX
//  9 : string + BX + CX + DX
// 10 : AL
// 11 : AL + DX
// 12 : AL + DSDX

//  new format:
//  %Rxx    xx = al,ah,bl,bh,cl,ch,dl,dh, ax,bx,cx,dx,si,di,bp,sp,ip,ds,es,cs,ss
//  %D      data
//  %S      Lascii string
//  %s      asciiZ string
//  %l      long
//  %L      long (NW order)
//  %w      word
//  %W      word (NW order)
//  %b      byte

struct RegisterTab {
  char *name;
  int  offset;
  int  size;
};

struct RegisterTab registerNameList[]={
 {  "ax", 0, 2},
 {  "bx", 2, 2},
 {  "cx", 4, 2},
 {  "dx", 6, 2},
 {  "ds", 8, 2},
 {  "si",10, 2},
 {  "es",12, 2},
 {  "di",14, 2},
 {  "ss",16, 2},
 {  "sp",18, 2},
 {  "bp",20, 2},
 {  "ip",22, 2},
 {  "cs",24, 2},
 {  "al", 0, 1},
 {  "ah", 1, 1},
 {  "bl", 2, 1},
 {  "bh", 3, 1},
 {  "cl", 4, 1},
 {  "ch", 5, 1},
 {  "dl", 6, 1},
 {  "dh", 7, 1},
 {  NULL,-1,-1}
};

char htoa(byte b)
{
  b=b&0xf;
  if (b<10)
    return b+'0';
  else
    return b-10+'a';
}

// returns printed value of register in <s>
char *LookupRegister(char *s, struct int_rec *r)
{
  struct RegisterTab *rt;
  static char   buf[20];
  byte   *dp;     // data ptr
  char   *bp;     // buf ptr
  int i;

  for (rt=registerNameList; rt->name ; rt++)
  {
    if (strcmp(rt->name, s)==0)
    {
      dp=((byte *)&(r->AX))+rt->offset;   // offset in recstruct
      bp=&buf[2*rt->size];                // offset in text buffer
      *bp--=0;                            // terminating 0
      for (i=0 ; i<rt->size ; i++)
      {
        *bp--=htoa(*dp);
        *bp--=htoa((*dp++)>>4);
      }
      return buf;
    }
  }
  return NULL;
}

void PrintString(byte *p, int n)
{
  bool outside=TRUE;
  int i;

  for (i=0 ; i<n ; i++)
    if (*p<' ')
    {
      if (!outside)
      {
        printf("\" ");
        outside=TRUE;
      }
      printf("%02x ",*p++);
    }
    else
    {
      if (outside)
      {
        printf("  \"");
        outside=FALSE;
      }
      putchar(*p++);
    }
  if (!outside) printf("\"");
}

char *getFCBname(char far *fcbfn)
{
  static char fn[13];
  int i;
  char *p;  char *q;
  p=fn;  q=p;
  for (i=0 ; i<8 ; i++)
  {
    *p=fcbfn[i];
    if (*p>0x20 && *p<0x7f && *p!='.')
      q=++p;  // remember end of name
  }
  *q++='.';

  p=q;
  for (i=0 ; i<3 ; i++)
  {
    *p=fcbfn[i];
    if (*p>0x20 && *p<0x7f && *p!='.')
      q=++p;  // remember end of name
  }
  *q++=0;
  return fn;
}

enum {
  PRT_NONE,
  PRT_CHARACTER,
  PRT_REGISTER,
  PRT_DATA
};

//
// process string with format characters
//   %Rxx or %Rexx  (xx=registername)
//   %Dm;n.          print n bytes data from data offset m (dump format)
//   %xm;n.          print n bytes hex from data offset m (intel byte order)
//   %Xm;n.          print n bytes hex from data offset m (moterola byte order)
//      (o=oct, b=bin, d=dec)
//   %sm;t;n.        print <t> terminated string from offset m, max n bytes
//   %Sm.            print length preceeded string from offset m
//   m == x[,y]  ,y == offset x after y-th string
//  [;n] is optional
//
int rprintf(char *fmt, struct int_rec *rec)
{
  char *p=fmt;
  char regName[5];
  bool percentDetected=FALSE;
  bool gettingRegisterName=FALSE;
  int  regNameIdx=0;
  int  printAction=PRT_NONE;
  int  dataPrinted=FALSE;
  char *regVal;

  while (*p)
  {
// process character
    printAction=PRT_NONE;
    if (percentDetected)
    {
      switch(*p)
      {
        case 'R':
          gettingRegisterName=TRUE;
          regNameIdx=0;
          break;
        case 'D':
          printAction=PRT_DATA;
          break;
        default:
          printf("!!! Invalid format character %c !!!\n", *p);
      }
      percentDetected=FALSE;
    }
    else if (gettingRegisterName)
    {
      regName[regNameIdx++]=*p;  regName[regNameIdx]=0;
      regVal=LookupRegister(regName, rec);
      if (regVal)
      {
        gettingRegisterName=FALSE;
        printAction=PRT_REGISTER;
      }
    }
    else if (*p=='%')
      percentDetected=TRUE;
    else
      printAction=PRT_CHARACTER;

// output actions
    switch (printAction)
    {
      case PRT_NONE:
        break;
      case PRT_CHARACTER:
        putchar(*p);
        break;
      case PRT_REGISTER:
        fputs(regVal, stdout);
        break;
      case PRT_DATA:
        PrintString((byte *)&rec[1], rec->reclen-sizeof(struct int_rec));
        dataPrinted=TRUE;
        break;
      default:
        printf("!!! error: invalid printAction %d !!!\n", printAction);
    }
    p++;
  }
  if (!dataPrinted)
  {
    putchar(' ');
    PrintString((byte *)&rec[1], rec->reclen-sizeof(struct int_rec));
  }
  return 0;
}

struct BTRVCallStruct {
  char far *BUF_OFFSET;
  int       BUF_LEN;
  char far *CUR_OFFSET;
  char far *FCB_OFFSET;
  int       FUNCTION;
  char far *KEY_OFFSET;
  char      KEY_LENGTH;
  char      KEY_NUMBER;
  int  far *STAT_OFFSET;
  int       XFACE_ID;
};

void PrintBtrv(byte *p, int n)
{
  struct BTRVCallStruct *btrv;
  btrv= (struct BTRVCallStruct *)p;
  printf("fn=%04d buf=%Fp(%d) cur=%Fp fcb=%Fp key=%Fp(%d):%d stat=%Fp id=%04x : ",
    btrv->FUNCTION, btrv->BUF_OFFSET, btrv->BUF_LEN,
    btrv->CUR_OFFSET, btrv->FCB_OFFSET,
    btrv->KEY_OFFSET, btrv->KEY_LENGTH, btrv->KEY_NUMBER,
    btrv->STAT_OFFSET, btrv->XFACE_ID);

  PrintString(p+sizeof(struct BTRVCallStruct), n-sizeof(struct BTRVCallStruct));
  putchar('\n');
}

/*
int textIDcmp(const void *key, const void *elem)
{
  if (((struct textID *)key)->id < ((struct textID *)elem)->id)
    return -1;
  else if (((struct textID *)key)->id > ((struct textID *)elem)->id)
    return 1;
  else
    return 0;
}
*/

struct textID *GetTextID(struct textID *tab, int id)
{
//  struct textID key;
//  key.id=id;
//  p=bsearch(&key, tab, tabsize, sizeof(struct textID), textIDcmp);

  struct textID *p;
  for (p=tab ; p->flag!=-1 ; p++)
    if (p->id==id)
      return p;
  return NULL;
}

void printrec(struct int_rec *rec)
{
  int  index;
  struct textID   *table;
  struct textID   *subTable;

  printf("%08lx ", rec->time);
  printf("%04x:%04x ", rec->CS, rec->IP);
  subTable=mainTable;

  do {
    if (subTable->flag & OFFSET)
      index=((byte *)&rec[1])[subTable->flag & 0xff];
    else if (subTable->flag & REG_BYTE)
      index=((byte *)&rec->AX)[ subTable->flag & 0xf];
    else if (subTable->flag & REG_WORD)
      index=((word *)&rec->AX)[ subTable->flag & 0xf];
    else if (subTable->flag & INTERRUPT_NUMBER)
      index=rec->intno;
    else
    {
      printf("Unknown geval\n");
      index=-1;
      break;
    }
    table=subTable;
    subTable=GetTextID(table->sub.table, index);
  } while (subTable && subTable->flag!=END);
  if (subTable==NULL)
  {
    rprintf(table->general, rec);
    printf(" [%02x]", index);
    if (rec->reclen>sizeof(*rec))
      rprintf(": %D", rec);
  }
  else
    rprintf(subTable->sub.format, rec);
  putchar('\n');
}

void usage(char *s, ...)
{
  va_list ap;
  va_start(ap,s);
  vprintf(s,ap);
  va_end(ap);
  printf("Usage: lstfile [-h] catchfile\n");
}

int main(int argc, char **argv)
{
  char *dmpname=NULL;
  int f;
  char data[MAXRECSIZE];
  struct int_rec *buf;
  long t0;
  int printtime=0;
  int record_length;

  while (argc>1)
  {
    if (argv[1][0]=='-')
    switch(argv[1][1])
    {
      case 'h':
          usage("Help screen\n");
          return 0;
      case 't':
          printtime++;
          break;
      default:
          usage("Invalid option: %s\n", argv[1]);
          return 1;
    }
    else if (dmpname==NULL)
      dmpname=argv[1];
    else
    {
      usage("Too many parameters: %s\n", argv[1]);
      return 1;
    }

    argc--; argv++;
  }
  if (dmpname==NULL)
  {
      usage("You have to specify catch dumpfile\n");
      return 1;
  }
  f=open(dmpname, O_RDONLY|O_BINARY);
  if (f<0)
  {
    perror(dmpname);
    return 1;
  }

  printf("                 time       CS   IP   data\n");
  t0=0;
  while (!eof(f))
  {
    read(f,data, sizeof(*buf));
    buf=(struct int_rec *)data;
    if (t0)
      printf("%8lx ", (buf->time)-t0);
    else
      printf("         ");
    t0=buf->time;
    record_length=buf->reclen;
    if (record_length==0) record_length=0x100;
    if (record_length>sizeof(*buf))
      read(f, data+sizeof(*buf), record_length-sizeof(*buf));
    printrec(buf);
  }
  close(f);
  return 0;
}
